
#ifndef PLS_SCHEDULER_H
#define PLS_SCHEDULER_H

#include <array>
#include <iostream>

#include "pls/internal/base/thread.h"
#include "pls/internal/base/barrier.h"
#include "pls/internal/scheduling/thread_state.h"
#include "root_master_task.h"
#include "root_worker_task.h"

namespace pls {
    namespace internal {
        namespace scheduling {
            // Upper thread limit for static memory allocation.
            // Could be moved to templating if needed.
            static constexpr int MAX_THREADS = 32;

            class scheduler {
                static void worker_routine() {
                    auto my_state = base::this_thread::state<thread_state>();

                    while (true) {
                        my_state->scheduler_->sync_barrier_.wait();
                        if (my_state->scheduler_->terminated_) {
                            return;
                        }

                        // The root task must only return when all work is done,
                        // because of this a simple call is enough to ensure the
                        // fork-join-section is done (logically joined back into our main thread).
                        my_state->root_task_->execute();

                        my_state->scheduler_->sync_barrier_.wait();
                    }
                }

                unsigned int num_threads_;
                std::array<thread_state, MAX_THREADS> thread_states_;
                std::array<base::thread<decltype(&worker_routine), thread_state>, MAX_THREADS> threads_;

                base::barrier sync_barrier_;
                bool terminated_;
            public:
                explicit scheduler(const unsigned int num_threads):
                    num_threads_{num_threads},
                    sync_barrier_{num_threads + 1},
                    terminated_{false} {
                    if (num_threads > MAX_THREADS) {
                        exit(1); // TODO: Exception Handling
                    }

                    for (unsigned int i = 0; i < num_threads; i++) {
                        thread_states_[i] = thread_state{this};
                        threads_[i] = base::start_thread(&worker_routine, &thread_states_[i]);
                    }
                }

                ~scheduler() {
                    terminate();
                }

                template<typename Function>
                void perform_work(Function work_section) {
                    root_master_task<Function> master{work_section};
                    root_worker_task<Function> worker{&master};

                    thread_states_[0].root_task_ = &master;
                    for (unsigned int i = 1; i < num_threads_; i++) {
                        thread_states_[i].root_task_ = &worker;
                    }

                    sync_barrier_.wait(); // Trigger threads to wake up
                    sync_barrier_.wait(); // Wait for threads to finish
                }

                void terminate(bool wait_for_workers=true) {
                    if (terminated_) {
                        return;
                    }

                    terminated_ = true;
                    sync_barrier_.wait();

                    if (wait_for_workers) {
                        for (unsigned int i = 0; i < num_threads_; i++) {
                            threads_[i].join();
                        }
                    }
                }
            };
        }
    }
}

#endif //PLS_SCHEDULER_H
